@@ -28,6 +28,20 @@ auth_callback (SoupAuthDomain *auth_domain, SoupMessage *msg,
 	return !strcmp (username, "user") && !strcmp (password, "password");
 }
 
+static void
+forget_close (SoupMessage *msg, gpointer user_data)
+{
+	soup_message_headers_remove (msg->response_headers, "Connection");
+}
+
+static void
+close_socket (SoupMessage *msg, gpointer user_data)
+{
+	SoupSocket *sock = user_data;
+
+	soup_socket_disconnect (sock);
+}
+
 static void
 server_callback (SoupServer *server, SoupMessage *msg,
 		 const char *path, GHashTable *query,
@@ -61,6 +75,41 @@ server_callback (SoupServer *server, SoupMessage *msg,
 		return;
 	}
 
+	if (g_str_has_prefix (path, "/content-length/")) {
+		gboolean too_long = strcmp (path, "/content-length/long") == 0;
+		gboolean no_close = strcmp (path, "/content-length/noclose") == 0;
+
+		soup_message_set_status (msg, SOUP_STATUS_OK);
+		soup_message_set_response (msg, "text/plain",
+					   SOUP_MEMORY_STATIC, "foobar", 6);
+		if (too_long)
+			soup_message_headers_set_content_length (msg->response_headers, 9);
+		soup_message_headers_append (msg->response_headers,
+					     "Connection", "close");
+
+		if (too_long) {
+			SoupSocket *sock;
+
+			/* soup-message-io will wait for us to add
+			 * another chunk after the first, to fill out
+			 * the declared Content-Length. Instead, we
+			 * forcibly close the socket at that point.
+			 */
+			sock = soup_client_context_get_socket (context);
+			g_signal_connect (msg, "wrote-chunk",
+					  G_CALLBACK (close_socket), sock);
+		} else if (no_close) {
+			/* Remove the 'Connection: close' after writing
+			 * the headers, so that when we check it after
+			 * writing the body, we'll think we aren't
+			 * supposed to close it.
+			 */
+			g_signal_connect (msg, "wrote-headers",
+					  G_CALLBACK (forget_close), NULL);
+		}
+		return;
+	}
+
 	soup_message_set_status (msg, SOUP_STATUS_OK);
 	if (!strcmp (uri->host, "foo")) {
 		soup_message_set_response (msg, "text/plain",
@@ -181,6 +230,7 @@ static void
 do_callback_unref_test (void)
 {
 	SoupServer *bad_server;
+	SoupAddress *addr;
 	SoupSession *session;
 	SoupMessage *one, *two;
 	GMainLoop *loop;
@@ -189,7 +239,12 @@ do_callback_unref_test (void)
 	debug_printf (1, "\nCallback unref handling\n");
 
 	/* Get a guaranteed-bad URI */
-	bad_server = soup_server_new (NULL, NULL);
+	addr = soup_address_new ("127.0.0.1", SOUP_ADDRESS_ANY_PORT);
+	soup_address_resolve_sync (addr, NULL);
+	bad_server = soup_server_new (SOUP_SERVER_INTERFACE, addr,
+				      NULL);
+	g_object_unref (addr);
+
 	bad_uri = g_strdup_printf ("http://127.0.0.1:%u/",
 				   soup_server_get_port (bad_server));
 	g_object_unref (bad_server);
@@ -317,6 +372,9 @@ do_msg_reuse_test (void)
 	g_free (signal_ids);
 }
 
+/* Server handlers for "*" work but are separate from handlers for
+ * all other URIs. #590751
+ */
 static void
 do_star_test (void)
 {
@@ -377,6 +435,19 @@ do_star_test (void)
 	soup_uri_free (star_uri);
 }
 
+/* Handle unexpectedly-early aborts. #596074, #618641 */
+static void
+ea_msg_completed_one (SoupSession *session, SoupMessage *msg, gpointer loop)
+{
+	debug_printf (2, "  Message 1 completed\n");
+	if (msg->status_code != SOUP_STATUS_CANCELLED) {
+		debug_printf (1, "  Unexpected status on Message 1: %d %s\n",
+			      msg->status_code, msg->reason_phrase);
+		errors++;
+	}
+	g_main_loop_quit (loop);
+}
+
 static gboolean
 ea_abort_session (gpointer session)
 {
@@ -411,16 +482,32 @@ do_early_abort_test (void)
 {
 	SoupSession *session;
 	SoupMessage *msg;
+	GMainContext *context;
+	GMainLoop *loop;
 
 	debug_printf (1, "\nAbort with pending connection\n");
 
 	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
+	msg = soup_message_new_from_uri ("GET", base_uri);
 
+	context = g_main_context_default ();
+	loop = g_main_loop_new (context, TRUE);
+	soup_session_queue_message (session, msg, ea_msg_completed_one, loop);
+	g_main_context_iteration (context, FALSE);
+
+	soup_session_abort (session);
+	while (g_main_context_pending (context))
+		g_main_context_iteration (context, FALSE);
+	g_main_loop_unref (loop);
+	soup_test_session_abort_unref (session);
+
+	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
 	msg = soup_message_new_from_uri ("GET", base_uri);
 
 	g_signal_connect (session, "connection-created",
 			  G_CALLBACK (ea_connection_created), NULL);
 	soup_session_send_message (session, msg);
+	debug_printf (2, "  Message 2 completed\n");
 
 	if (msg->status_code != SOUP_STATUS_CANCELLED) {
 		debug_printf (1, "    Unexpected response: %d %s\n",
@@ -429,6 +516,68 @@ do_early_abort_test (void)
 	}
 	g_object_unref (msg);
 
+	while (g_main_context_pending (context))
+		g_main_context_iteration (context, FALSE);
+
+	soup_test_session_abort_unref (session);
+}
+
+static void
+do_content_length_framing_test (void)
+{
+	SoupSession *session;
+	SoupMessage *msg;
+	SoupURI *request_uri;
+	goffset declared_length;
+
+	debug_printf (1, "\nInvalid Content-Length framing tests\n");
+
+	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC, NULL);
+
+	debug_printf (1, "  Content-Length larger than message body length\n");
+	request_uri = soup_uri_new_with_base (base_uri, "/content-length/long");
+	msg = soup_message_new_from_uri ("GET", request_uri);
+	soup_session_send_message (session, msg);
+	if (msg->status_code != SOUP_STATUS_OK) {
+		debug_printf (1, "    Unexpected response: %d %s\n",
+			      msg->status_code, msg->reason_phrase);
+		errors++;
+	} else {
+		declared_length = soup_message_headers_get_content_length (msg->response_headers);
+		debug_printf (2, "    Content-Length: %lu, body: %s\n",
+			      (gulong)declared_length, msg->response_body->data);
+		if (msg->response_body->length >= declared_length) {
+			debug_printf (1, "    Body length %lu >= declared length %lu\n",
+				      (gulong)msg->response_body->length,
+				      (gulong)declared_length);
+			errors++;
+		}
+	}
+	soup_uri_free (request_uri);
+	g_object_unref (msg);
+
+	debug_printf (1, "  Server claims 'Connection: close' but doesn't\n");
+	request_uri = soup_uri_new_with_base (base_uri, "/content-length/noclose");
+	msg = soup_message_new_from_uri ("GET", request_uri);
+	soup_session_send_message (session, msg);
+	if (msg->status_code != SOUP_STATUS_OK) {
+		debug_printf (1, "    Unexpected response: %d %s\n",
+			      msg->status_code, msg->reason_phrase);
+		errors++;
+	} else {
+		declared_length = soup_message_headers_get_content_length (msg->response_headers);
+		debug_printf (2, "    Content-Length: %lu, body: %s\n",
+			      (gulong)declared_length, msg->response_body->data);
+		if (msg->response_body->length != declared_length) {
+			debug_printf (1, "    Body length %lu != declared length %lu\n",
+				      (gulong)msg->response_body->length,
+				      (gulong)declared_length);
+			errors++;
+		}
+	}
+	soup_uri_free (request_uri);
+	g_object_unref (msg);
+
 	soup_test_session_abort_unref (session);
 }
 
@@ -457,8 +606,10 @@ main (int argc, char **argv)
 	do_msg_reuse_test ();
 	do_star_test ();
 	do_early_abort_test ();
+	do_content_length_framing_test ();
 
 	soup_uri_free (base_uri);
+	soup_test_server_quit_unref (server);
 
 	test_cleanup ();
 	return errors != 0;
